from mupixutils import *
import argparse
import json
import os

def main():
    # Create an ArgumentParser object
    parser = argparse.ArgumentParser(description='Train µPIX Model')

    # Add arguments for the input and output paths
    parser.add_argument('--experiment_path', type=str, required=True, help='Path to the experiment folder')

    # Add an optional argument for retraining
    parser.add_argument('--retrain', action='store_true', help='Set this flag for retraining')


    # Parse the arguments
    args = parser.parse_args()

    # Access the paths and retrain flag from the arguments
    experiment_path = args.experiment_path
    retrain = args.retrain

    # Load the JSON file
    with open(experiment_path+"/hyperparameters.json", 'r') as file:
        config = json.load(file)

    # Accessing the loaded configuration data
    learning_rate_generator = config["learning_rate_generator"]
    learning_rate_discriminator = config["learning_rate_discriminator"]
    batch_size = config["batch_size"]
    num_epochs = config["num_epochs"]
    loss_weight = config["loss_weight"]
    tile_size = config["tile_size"]
    patience = config["patience"]
    valid_size = config["valid_size"]
    seed = config["seed"]

    # Access data paths
    clean_data_path = config["data_paths"]["clean"]
    noisy_data_path = config["data_paths"]["noisy"]

    # Validate that model_path is provided if retraining is True
    if retrain and len(os.listdir(experiment_path+"/results/networks")) == 0:
        raise ValueError("Retraining is enabled but no network has been found.")
    output_path = experiment_path +"/results"
    # Output for verification
    print(f"Input Path: {noisy_data_path,clean_data_path}")
    print(f"Output Path: {output_path}")
    print(f"Retraining: {retrain}")

    print("Data loading...")
    train_dataset, val_dataset = dataloading(noisy_data_path,clean_data_path,valid_size,tile_size,seed)
    image_shape = train_dataset[0].shape[1:]
    print("Data loaded !")
    
    if retrain:
        model_path = experiment_path+"/results/networks"
        print(f"Model Path: {model_path}")
        if not os.path.exists(model_path):
            raise FileNotFoundError(f"Model path {model_path} does not exist.")
        print("Retraining the model with the previous weights...")
        g_model = tf.keras.models.load_model(model_path+"/Generator")
        d_model = tf.keras.models.load_model(model_path+"/Discriminator") 
        gan_model = define_gan(g_model, d_model, image_shape)
        last_epoch=last_epoch_searcher(output_path+"/images")

    else:
        print("Training from scratch...")
        d_model = define_discriminator(image_shape)
        g_model = define_generator(image_shape)
        gan_model = define_gan(g_model, d_model, image_shape)
        last_epoch=0

    gan_model.compile(loss=['binary_crossentropy', 'mse'], optimizer=Adam(learning_rate=learning_rate_generator, beta_1=0.8) , loss_weights=[1,loss_weight])
    g_model.compile(loss='mse', optimizer=Adam(learning_rate=learning_rate_generator, beta_1=0.8))
    d_model.compile(loss='binary_crossentropy', optimizer=Adam(learning_rate=learning_rate_discriminator, beta_1=0.9))
    # history is will have mse of train, mse of valid and ssim of valid
    history = train(d_model, g_model, gan_model, train_dataset, output_path,
        n_epochs=num_epochs, batch_size=batch_size, val_dataset=val_dataset, patience = patience, starting_epoch=last_epoch) #training loop



if __name__ == "__main__":
    main()

