import gc
import os
import re
import json

os.environ["SM_FRAMEWORK"] = "tf.keras"
from os import listdir
import numpy as np # type: ignore
from numpy import zeros, ones # type: ignore
from tifffile import imread, imwrite # type: ignore
import cv2 # type: ignore
import albumentations as A # type: ignore
import tensorflow as tf # type: ignore
from tensorflow.keras.optimizers import Adam # type: ignore
from tensorflow.keras.utils import Sequence # type: ignore
from tensorflow.image import ssim # type: ignore
from keras import backend as K # type: ignore
from keras.initializers import RandomNormal # type: ignore
from keras.models import Model # type: ignore
from keras.layers import ( # type: ignore
    Input, Conv2D, LeakyReLU, Activation, 
    Concatenate, BatchNormalization)
from tensorflow import keras # type: ignore
import segmentation_models as sm # type: ignore
from segmentation_models import Unet, get_preprocessing # type: ignore
from sklearn.model_selection import train_test_split # type: ignore



AUGMENTATIONS_TRAIN =A.Compose([
    A.ShiftScaleRotate (p=.25,border_mode = cv2.BORDER_REFLECT_101,),
    A.ElasticTransform (p=.25),
    A.OpticalDistortion(p=.25),
    A.RandomRotate90(p=.75),
    A.HorizontalFlip(p=.5)
    ])

class Imgen(Sequence):

    def __init__(self, x_set, y_set, batch_size, augmentations):
        self.x, self.y = x_set, y_set
        self.batch_size = batch_size
        self.augment = augmentations

    def __len__(self):
        return int(np.ceil(len(self.x) / float(self.batch_size)))

    def __getitem__(self, idx):
        batch_x = np.copy(self.x[idx * self.batch_size:(idx + 1) * self.batch_size])
        batch_y = np.copy(self.y[idx * self.batch_size:(idx + 1) * self.batch_size])
        img=[]
        msk=[]
        for i in range(batch_x.shape[0]):
            agm = self.augment(image=batch_x[i], mask=batch_y[i])
            img.append(agm["image"])
            msk.append(agm["mask"])
        img=np.array(img)
        msk=np.array(msk)
        return np.array((img*2)-1).astype("float32"),np.array((msk*2)-1).astype("float32")
"""
Defining the functions to create the µPIX network inspired by Jason Brownlee's Pix2Pix Example
"""
# define the discriminator model
def define_discriminator(image_shape):
    # weight initialization
    init = RandomNormal(stddev=0.02) #( RandomNormal is a initializer that generates tensors with a normal distribution.)
    # source image input
    in_src_image = Input(shape=image_shape)
    # target image input
    in_target_image = Input(shape=image_shape)
    # concatenate images channel-wise
    merged = Concatenate()([in_src_image, in_target_image])
    # C64
    d = Conv2D(64, (3,3), strides=(2,2), padding='same', kernel_initializer=init)(merged)
    d = LeakyReLU(alpha=0.2)(d)
    # C128
    d = Conv2D(128, (3,3), strides=(2,2), padding='same', kernel_initializer=init)(d)
    d = BatchNormalization()(d)
    d = LeakyReLU(alpha=0.2)(d)
    # C256
    d = Conv2D(256, (3,3), strides=(2,2), padding='same', kernel_initializer=init)(d)
    d = BatchNormalization()(d)
    d = LeakyReLU(alpha=0.2)(d)
    # C512
    d = Conv2D(512, (3,3), strides=(2,2), padding='same', kernel_initializer=init)(d)
    d = BatchNormalization()(d)
    d = LeakyReLU(alpha=0.2)(d)
    # second last output layer
    d = Conv2D(512, (3,3), padding='same', kernel_initializer=init)(d)
    d = BatchNormalization()(d)
    d = LeakyReLU(alpha=0.2)(d)
    # patch output
    d = Conv2D(1, (3,3), padding='same', kernel_initializer=init)(d)
    patch_out = Activation('sigmoid')(d)
    # define model
    model = Model([in_src_image, in_target_image], patch_out)
    # compile model
    opt = Adam(learning_rate=1e-4, beta_1=0.8)   
    model.compile(loss='binary_crossentropy', optimizer=opt)
    return model

# define the standalone generator model
def define_generator(image_shape=(256,256,1)):
    # define model
    sm.set_framework('tf.keras')
    sm.framework()
    model = Unet("efficientnetb0",encoder_weights=None,activation="tanh", input_shape=image_shape,classes=1)
    return model

# define the combined generator and discriminator model, for updating the generator
def define_gan(g_model, d_model, image_shape):
    # make weights in the discriminator not trainable
    for layer in d_model.layers:
        if not isinstance(layer, BatchNormalization):
            layer.trainable = False 
    # define the source image
    in_src = Input(shape=image_shape)
    # connect the source image to the generator input
    gen_out = g_model(in_src)
    # connect the source input and generator output to the discriminator input
    dis_out = d_model([in_src, gen_out])
    # src image as input, generated image and classification output
    model = Model(in_src, [dis_out, gen_out])
    # compile model
    opt = Adam(learning_rate=1e-4, beta_1=0.8)
    model.compile(loss=['binary_crossentropy', 'mse'], optimizer=opt, loss_weights=[1,10])
    return model

# select a batch of random samples, returns images and target
def generate_real_samples(subdataset, n_samples, patch_shape):
    # retrieve selected images
    X1, X2 = subdataset
    # generate 'real' class labels (1)
    y = ones((n_samples, patch_shape, patch_shape, 1))
    return [X1[...,None], X2[...,None]], y

# generate a batch of images, returns images and targets
def generate_fake_samples(g_model, samples, patch_shape):
    # generate fake instance
    X = g_model.predict(np.squeeze(samples)[...,None])
    # create 'fake' class labels (0)
    y = zeros((len(X), patch_shape, patch_shape, 1))
    return X, y

def mean_squared_error(y_true, y_pred):
    return tf.reduce_mean(tf.square(y_true - y_pred))

def structural_similarity_index(y_true, y_pred, max_val=1.0):
    return tf.reduce_mean(ssim(y_true, y_pred, max_val=max_val))

def train(d_model, g_model, gan_model, dataset, output_path, val_dataset = None, n_epochs=100, batch_size=16, patience = 20, starting_epoch=0):

    # determine the output square shape of the discriminator
    n_patch = d_model.output_shape[1]
    # calculate the number of batches per training epoch
    bat_per_epo = int(len(dataset[0]) / batch_size)
    # calculate the number of training iterations
    n_steps = bat_per_epo * n_epochs
    j=None
    train_gen=None
    best_val_loss = float('inf')
    cur_patience=patience
    current_epoch = 0 + starting_epoch
    history = [],[],[]
    if val_dataset is not None : xval,yval = val_dataset[0],val_dataset[1]
    for i in range(n_steps):
        # select a batch of real samples
        if j is None or j==bat_per_epo-1:
            del train_gen
            gc.collect()
            j=0 # generating a dataset shuffling
            current_epoch+=1
            tab_ix = np.random.permutation(len(dataset[0]))
            # cutting it into shuffled batches
            trainA, trainB = dataset[0][tab_ix],dataset[1][tab_ix]
            train_gen = Imgen(trainA, trainB, batch_size=batch_size, augmentations=AUGMENTATIONS_TRAIN)
            del trainA, trainB, tab_ix
            # val loss computation
            if val_dataset is not None :
                # print(np.mean(xval[0]),np.mean(yval[0]))
                v_res = (g_model.predict((xval/127.5)-1)+1)*127.5
                mse=(mean_squared_error(v_res,yval)).numpy()
                ssim=((structural_similarity_index(v_res,yval)).numpy()) 
                history[1].append(mse)
                history[2].append(ssim)
                print("Epoch", current_epoch)
                # print('Val> mse[%.3e], ssim[%.3e]' % (mse, ssim))
                print("Val> mse[",str(mse),"], ssim[",str(ssim),"]")
                cur_patience-=1
                val_loss = mse/((ssim+1)/2) # scaling ssim on 0,1 instead of -1,1 : trouble when ssim is <0
                with open(output_path+"/log.txt", 'a') as file:
                        # file.write("[Loss] Val loss at epoch %d : MSE[%.3e], SSIM[%.3e], Validation_Loss[%.3e]\n"%(current_epoch,mse,ssim,val_loss))
                        file.write("[Loss] Val loss at epoch "+str(current_epoch)+" : MSE["+str(mse)+"], SSIM["+str(ssim)+"], Validation_Loss["+str(val_loss)+"]\n")
                if val_loss < best_val_loss:
                    best_val_loss = np.copy(val_loss)
                    g_model.save(output_path+"/networks/Generator")
                    d_model.save(output_path+"/networks/Discriminator")
                    imwrite(output_path+"/images/Validation_epoch_%d.tif"%(current_epoch),np.concatenate((xval[0],v_res[0],yval[0]),axis=1))
                    cur_patience = patience
                else:
                    with open(output_path+"/log.txt", 'a') as file:
                            file.write("Loss did not improve : best val_loss was: "+str(best_val_loss)+" and current val_loss is "+str(val_loss)+"\n")

                if cur_patience == 0 : 
                    print("Patience has been reached, training stopped")
                    if g_model.optimizer.learning_rate.numpy()>1e-7:
                        print("Training stuck, lowering learning rate :" + str(g_model.optimizer.learning_rate.numpy()) + " => " + str(g_model.optimizer.learning_rate.numpy()*.1))
                        with open(output_path+"/log.txt", 'a') as file:
                            file.write("Training stuck, lowering learning rate :" + str(g_model.optimizer.learning_rate.numpy()) + " => " + str(g_model.optimizer.learning_rate.numpy()*.1)+"\n")

                        K.set_value(g_model.optimizer.learning_rate, g_model.optimizer.learning_rate.numpy()*.1)
                        K.set_value(d_model.optimizer.learning_rate, d_model.optimizer.learning_rate.numpy()*.1)
                        cur_patience=patience
                    else :
                        with open(output_path+"/log.txt", 'a') as file:
                            # file.write("[Patience]: Reached at epoch %d with best validation loss :[%.3e]\n"%(current_epoch,val_loss))
                            file.write("[Patience]: Reached at epoch "+str(current_epoch)+"with best validation loss :["+str(val_loss)+"]\n")
                        break

            gc.collect()
        
        # ix holds the id of the images of the permutated batch n°j
        sampleX, sampleY = train_gen.__getitem__(j)
        # next permutation index (next batch)
        j+=1
        [X_realA, X_realB], y_real = generate_real_samples((sampleX,sampleY), batch_size, n_patch)
        del sampleX, sampleY
        gc.collect()
        # generate a batch of fake samples
        X_fakeB, y_fake = generate_fake_samples(g_model, X_realA, n_patch)
        # update discriminator for real samples
        _ = d_model.train_on_batch([X_realA, X_realB], y_real)
        # update discriminator for generated samples
        _ = d_model.train_on_batch([X_realA, X_fakeB], y_fake)
        # update the generator
        g_loss1,_ ,_ = gan_model.train_on_batch(X_realA, [y_real, X_realB])
        # summarize performance
        # print('Step>%d, Generator loss : %.3e' % (j, g_loss1))
        print('Step>',str(j),', Generator loss : ', str(g_loss1))
        # Saving the network every time it become better and when the network already did at least 500 batches.
        history[0].append(g_loss1)
        del _,g_loss1
        gc.collect()
    return history


class Tile:
    '''
    Usage    : Tile(image, tilesize, overlap, verbose)
    image    : Array of shape [height, width, channels]
    tilesize : Output shape will be [tilesize, tilesize, channels]
    overlap  : Pixels common between two consecutive tiles
    =============================================================
    Methods  :
    padding(mode="reflect") : Returns padded image
    tilegeneration()        : Returns a list of image tiles
    '''

    def __init__(self, image, tilesize=None, overlap=None, verbose=False):
        self.im = image
        self.s = np.shape(self.im)[:2]
        self.t = tilesize or self.s[0] // 4
        self.o = overlap or self.t // 4
        self.verbose = verbose
        self.to = self.t - self.o
        self.nbt = [(self.s[0] + self.o) // self.to, (self.s[1] + self.o) // self.to]
        self.pIm = None
        self._validate()

    def _validate(self):
        if self.s[0] % 2: self.im = self.im[:-1, :]
        if self.s[1] % 2: self.im = self.im[:, :-1]
        if self.o % 2: self.o -= 1
        if self.t > min(self.s) and self.verbose: print("Tile size is larger than image.")

    def _detpad(self):
        psize = [(self.nbt[i] * self.to + self.o) - self.s[i] for i in range(2)]
        return [(p // 2, p - p // 2) for p in psize]

    def padding(self, mode="reflect"):
        if self.pIm is None:
            pad_h, pad_w = self._detpad()
            pad = ((pad_h[0], pad_h[1]), (pad_w[0], pad_w[1]), (0, 0))
            self.pIm = np.pad(self.im, pad, mode=mode)
            gc.collect()
        return self.pIm

    def tilegeneration(self):
        self.padding()
        tiles = []
        for i in range(self.nbt[0]):
            for j in range(self.nbt[1]):
                tile = self.pIm[i*self.to:i*self.to+self.t, j*self.to:j*self.to+self.t]
                tiles.append(tile)
        return tiles

"""Image Processing for network training"""
#calling of the tiling procedure and defining a function to apply on a whole tensor
def tiling(dataset,tilesize=256,overlap=64):
    
    tiled=[]
    for i in dataset:
        tempi=Tile(i,tilesize,overlap)
        tiled.append(tempi.tilegeneration())
    del tempi
    gc.collect()
    return np.reshape(np.concatenate(tiled,axis=0),(-1,tilesize,tilesize,1))

def percentile_normalize(image, lower_percentile=1, upper_percentile=99):
    """
    Normalise l'image en utilisant la normalisation par centile.
    
    :param image: Tableau NumPy représentant l'image
    :param lower_percentile: Percentile inférieur pour la normalisation (par défaut: 1)
    :param upper_percentile: Percentile supérieur pour la normalisation (par défaut: 99)
    :return: Image normalisée
    """
    # Calcul des percentiles
    lower_bound = np.percentile(image, lower_percentile)
    upper_bound = np.percentile(image, upper_percentile)
    
    # Normalisation
    normalized_image = np.clip(image, lower_bound, upper_bound)
    normalized_image = 1* (normalized_image - lower_bound) / (upper_bound - lower_bound)
    
    return normalized_image

def percentile_normalize_test(image, lower_percentile=1, upper_percentile=99):
    """
    Normalise l'image en utilisant la normalisation par centile.
    
    :param image: Tableau NumPy représentant l'image
    :param lower_percentile: Percentile inférieur pour la normalisation (par défaut: 1)
    :param upper_percentile: Percentile supérieur pour la normalisation (par défaut: 99)
    :return: Image normalisée
    """
    # Calcul des percentiles
    lower_bound = np.percentile(image, lower_percentile)
    upper_bound = np.percentile(image, upper_percentile)
    
    # Normalisation
    normalized_image = np.clip(image, lower_bound, upper_bound)
    normalized_image = 255* (normalized_image - lower_bound) / (upper_bound - lower_bound)
    
    return normalized_image

# 'normalized_image' contiendra l'image normalisée avec des valeurs entre 0 et 1

def dataloading(pathx,pathy, val_size=0.1,tilesize=256,seed=42):
    # Load images and handle inhomogeneous sizes
    datax = [np.squeeze(imread(pathx + "/" + im))[...,None] for im in sorted(listdir(pathx))]
    datay = [np.squeeze(imread(pathy + "/" + im))[...,None] for im in sorted(listdir(pathy))]
    print(len(datax), "images loaded")
    # Split the dataset into training and validation sets
    x_train, x_val, y_train, y_val = train_test_split(datax, datay, test_size=val_size, random_state=seed)
    

    # Tiling and normalization for training data
    x_train_tiled = tiling(x_train,tilesize=tilesize)
    y_train_tiled = tiling(y_train,tilesize=tilesize)

    print(len(x_train_tiled), "tiles extracted for training")
    x_train_processed = percentile_normalize(np.array(x_train_tiled)).astype("float32")
    y_train_processed = percentile_normalize(np.array(y_train_tiled)).astype("float32")
    train_dataset = [x_train_processed, y_train_processed]


    # Tiling and normalization for training data
    x_val_tiled = tiling(x_val,tilesize=tilesize)
    y_val_tiled = tiling(y_val,tilesize=tilesize)

    print(len(x_val_tiled), "tiles extracted for validation")
    x_val_processed = 255 * percentile_normalize(np.array(x_val_tiled)).astype("float32")
    y_val_processed = 255 * percentile_normalize(np.array(y_val_tiled)).astype("float32")
    val_dataset = [x_val_processed, y_val_processed]
    return train_dataset, val_dataset

# def test_dataloading(test_data_path, experiment_path, batch_size):
#     # Load images and keep track of filenames
#     image_filenames = sorted(os.listdir(test_data_path))  # Sorted list of filenames
#     data = [imread(os.path.join(test_data_path, im)) for im in image_filenames]
    
#     # Convert images to tensor
#     data_stack = np.stack(data, axis=0)
#     data_tensor = percentile_normalize(data_stack)
#     print(len(data_tensor), "images loaded")

#     # Load the model
#     g_model = tf.keras.models.load_model(os.path.join(experiment_path, "results/networks/Generator"))

#     # Check if the weights file exists before saving
#     weights_file = os.path.join(experiment_path, "results/networks/Generator.h5")
#     g_model.save_weights(weights_file)
#     print(f"Saved weights to {weights_file}")
    
#     # Load the model for predictions
#     pred_model = Unet("efficientnetb0", encoder_weights=None, activation="tanh", input_shape=(None, None, 1), classes=1)
#     pred_model.load_weights(weights_file)
    
#     # Predict
#     predictions = (pred_model.predict((data_tensor[..., None] / 127.5) - 1, batch_size=batch_size) + 1) * 127.5

#     # Create the output directory if it doesn't exist
#     output_dir = os.path.join(experiment_path, "predictions")
#     os.makedirs(output_dir, exist_ok=True)

#     # Save each predicted image with the corresponding filename
#     for filename, pred in zip(image_filenames, predictions):
#         output_path = os.path.join(output_dir, filename)  # Preserve filename
#         imwrite(output_path, pred)
#         print(f"Saved prediction: {output_path}")

#     print("All predictions saved successfully!")

def test_dataloading(test_data_path, experiment_path, batch_size):
    # Load model
    g_model = tf.keras.models.load_model(os.path.join(experiment_path, "results/networks/Generator"))
    
    # Save weights
    weights_file = os.path.join(experiment_path, "results/networks/Generator.h5")
    g_model.save_weights(weights_file)
    print(f"Saved weights to {weights_file}")
    
    # Load model for predictions
    pred_model = Unet("efficientnetb0", encoder_weights=None, activation="tanh", input_shape=(None, None, 1), classes=1)
    pred_model.load_weights(weights_file)
    
    # Create output directory
    output_dir = os.path.join(experiment_path, "predictions")
    os.makedirs(output_dir, exist_ok=True)
    
    # Process images one by one
    image_filenames = sorted(os.listdir(test_data_path))
    for filename in image_filenames:
        image_path = os.path.join(test_data_path, filename)
        image = imread(image_path)
        image_tensor = percentile_normalize_test(image)[..., None]  # Normalize and add channel
        image_tensor = ((image_tensor / 127.5) - 1)[None, ...]  # Prepare for model
        
        # Predict
        prediction = (pred_model.predict(image_tensor, batch_size=1)[0] + 1) * 127.5
        
        # Save output
        output_path = os.path.join(output_dir, filename)
        imwrite(output_path, prediction)
        print(f"Saved prediction: {output_path}")
    
    print("All predictions saved successfully!")


def last_epoch_searcher(folder):

    # List to hold the epoch numbers
    epochs = []

    # Loop through each file in the folder
    for filename in os.listdir(folder):
        # Use regex to find the epoch number in the filename
        match = re.search(r'Validation_epoch_(\d+)\.tif', filename)
        if match:
            # Append the epoch number as an integer
            epochs.append(int(match.group(1)))

    # Get the maximum epoch number
    if epochs:
        return max(epochs)
    else:
        return 0